/*
GBPrinter/PC interface v1.0
CC BY furrtek 2009
http://www.furrtek.org
Atmega8 16Mhz
*/
#define F_CPU 16000000UL

#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <inttypes.h>
#include <avr/interrupt.h>

// Pin defs relative to gbprinter serial port:
#define		SCK		PC1
#define		SOUT	PC2
#define		SIN		PC3

uint8_t Presence;
uint8_t volatile fail;

void SerialPut(uint8_t byte) {
	while((UCSRA&(1<<UDRE)) == 0);
	UDR = byte;
}

ISR(TIMER1_OVF_vect) {
	SerialPut('i');
	TCCR1B = 0b00000000;
	fail=1;
}

uint8_t GBSerialOut(uint8_t command) {
	uint8_t clock,reply=0;
	for (clock=0;clock<8;clock++) {
		reply<<=1;					// Shifte la rponse
		PORTC &= ~_BV(SCK);
		if ((command << clock) & 0x80)
			PORTC |= _BV(SOUT);
		else
			PORTC &= ~_BV(SOUT);
		_delay_us(60);				// Bas pendant 60s
		PORTC |= _BV(SCK);
		if (bit_is_set(PINC,SIN)) reply |= 1;
		_delay_us(60);				// Haut pendant 60s
	}
	_delay_us(175-60);				// 175s entre deux octets
	return reply;
}

void CRLF(void) {
	while((UCSRA&(1<<UDRE)) == 0);
	UDR = 0x0D; // CR, LF
	_delay_ms (10);
	while((UCSRA&(1<<UDRE)) == 0);
	UDR = 0x0A;
	_delay_ms (10);
}

void SerialBinary(uint8_t a) {
	for (uint8_t bit=0x00;bit<8;bit++) {
		while((UCSRA&(1<<UDRE)) == 0);
		if ((a << bit) & 0x80) {
			UDR = 0x31; // "1"
		} else {
			UDR = 0x30; // "0"
		}
	}
}

void GBPCommand(uint8_t command, uint8_t arg1, uint8_t arg2, uint8_t arg3) {
	GBSerialOut(0x88);				// Magic Bytes
	GBSerialOut(0x33);
	GBSerialOut(command);			// Commande
	GBSerialOut(arg1);				// Arguments...
	GBSerialOut(arg2);
	GBSerialOut(arg3);
}

uint8_t GBPFooter(uint8_t command, uint8_t arg1, uint8_t arg2, uint8_t arg3, uint16_t CRC) {
	CRC=command+arg1+arg2+arg3+CRC;
	GBSerialOut((uint8_t)CRC & 0xFF);			// CRC Low
	GBSerialOut((CRC & 0xFF00)>>8);				// CRC High
	Presence = GBSerialOut(0x00);				// Devrait tre 129
	return GBSerialOut(0x00);					// Status
}

int main(void) {
	uint8_t Status,bufferindex;
	uint16_t CRCData,packet;
	uint8_t buffer[0x280];
	WDTCR = (1<<WDCE) | (1<<WDE);	// Kick doggie
	WDTCR = 0x00;

	DDRC = 0b00000110;				// SCK et SOUT en outputs
	PORTC = 0b00000000;				// Pas de pull-ups

	TIMSK = 0b00000100; 			// Interruption sur overflow TMR1

    #define BAUD 38400
    #include <util/setbaud.h>
    UBRRH = UBRRH_VALUE;
    UBRRL = UBRRL_VALUE;
    #if USE_2X
    UCSRA |= (1 << U2X);
    #else
    UCSRA &= ~(1 << U2X);
    #endif
	UCSRC = (1<<URSEL)|(1<<UCSZ1)|(1<<UCSZ0);
	UCSRB = (1<<TXEN)|(1<<RXEN);

	sei();

	while(1) {
		do {
			SerialPut('D');
			_delay_ms(500);
			while(!(UCSRA&(1<<RXC))) {};
		} while (UDR != 'A');

		// Commande 1, attend 0x81
		while(1) {
			GBPCommand(0x01,0x00,0x00,0x00);
			Status = GBPFooter(0x01,0x00,0x00,0x00,0x0000);
			if (Presence != 129) {
				_delay_ms(1000);
			} else {
				break;
			}
		}
		SerialPut('B');

		do {
		TCCR1B = 0b00000101;
		fail=0;
		for (packet=0x0000;packet<0x0280;packet++) {
			while(!(UCSRA&(1<<RXC))){
				if (fail) break;
			};
			buffer[packet]=UDR;
			TCNT1=0;
			// Bit 0 de status = CRC Fail ?
			// Bit 3 de status = Prt  imprimer une bande (16 lignes) ?
		}
		TCCR1B = 0b00000000;
		} while (fail);

		// Commande 4
		GBPCommand(0x04,0x00,0x80,0x02);
		CRCData=0;
		for (packet=0;packet<0x0280;packet++) {
			GBSerialOut(buffer[packet]);
			CRCData += buffer[packet];
		}
		Status = GBPFooter(0x04,0x00,0x80,0x02,CRCData);

		// Commande 4, 0 (terminateur)
		GBPCommand(0x04,0x00,0x00,0x00);
		Status = GBPFooter(0x04,0x00,0x00,0x00,0x00);

		// Commande 2, 0 marge, 3 marges, palette normale, exposition 0x40
		GBPCommand(0x02,0x00,0x04,0x00);
		GBSerialOut(0x01);
		GBSerialOut(0x00);
		GBSerialOut(0xE4);
		GBSerialOut(0x40);
		Status = GBPFooter(0x02,0x00,0x04,0x00,0x01+0x00+0xE4+0x40);

		// Commande 15, attend que le bit 1 de Status passe  1 (impression)
		while(1) {
			GBPCommand(0x0F,0x00,0x00,0x00);
			Status = GBPFooter(0x0F,0x00,0x00,0x00,0x00);
			if (bit_is_set(Status,1)) break;
		}

		// Commande 15, attend que le bit 1 de Status passe  0 (Termin)
		while(1) {
			GBPCommand(0x0F,0x00,0x00,0x00);
			Status = GBPFooter(0x0F,0x00,0x00,0x00,0x00);
			if (!bit_is_set(Status,1)) break;
			_delay_ms(500);
		}
	}

    return 0;			// Obligatoire ?
}
